#include "rstream.h"

#define LIMPAR 20e6  // in bytes

const unsigned long FACTOR = 48271;

const unsigned long max_value = 2147483647; // (2**(31))-1

// This is used in the poisson function
const float cof[6] = {76.18009173, -86.50532033, 24.01409822,
                      -1.231739516, 0.120858003e-2, -0.536382e-5};


/***********************************************************************
** function:    Rstream, constructor;
** parameters:  unsigned long int, a seed s;
** description: Create a stream with a seed s.
***********************************************************************/
Rstream::Rstream(unsigned long int s) {

  initStream(s);

} // End constructor


/***********************************************************************
** function:    initStream;
** parameters:  unsigned long int, a seed s;
** description: Initialize a stream with a seed s.
***********************************************************************/
void Rstream::initStream(unsigned long int s) {

  xi = (s % 2)?s:(s+1); // Make sure the seed is an odd number

} // End initStream


/***********************************************************************
** function:    uniform01;
** parameters:  none;
** description: Generate a random number in the open interval (0,1),
** based on the relation
**               X(i+1)  =  (X(i) * 48271) mod (2**31-1).
***********************************************************************/
double Rstream::uniform01() {

  xi *= FACTOR;
  xi %= max_value;
  return ((double)xi / (double)max_value);

} // End uniform01()


/***********************************************************************
** function:    get_xi;
** parameters:  none;
** description: gets the current value of xi.
***********************************************************************/
const unsigned long int Rstream::get_xi() const {

  return xi;

} // End get_xi


/***********************************************************************
** function:    uniform;
** parameters:  double, lower and upper;
** description: Generate a random number between (lower,upper)
***********************************************************************/
double Rstream::uniform(double lower, double upper) {

  return (uniform01() * (upper - lower) + lower);

} // End uniform


/***********************************************************************
** function:    UniformLong
** parameters:  long, lower and upper;
** description: Generate a random integer from [lower,upper]
***********************************************************************/
long Rstream::UniformLong(long lower, long upper) {
  return( lower+ ((long) floor(uniform01()*(upper-lower+1))));
}

/***********************************************************************
** function:    floor_uniform;
** parameters:  double, lower and upper;
** description: Generate a random integer between lower and (upper - 1).
***********************************************************************/
double Rstream::floor_uniform(double lower, double upper) {

  return floor(uniform(lower,upper));

} // End floor_uniform


/***********************************************************************
** function:    Exponential;
** parameters:  double rate
** description: Generate an exponentially distributed random with rate
**              exponential distribution. The value returned 
**              represents the interval between events occurring at
**              the supplied rate (rate = 1/mean).
***********************************************************************/
double Rstream::Exponential(double rate) {

  return (double) log(uniform01())/(-1.0*rate);

} // End floor_uniform

/***********************************************************************
** function:    lngamma;
** parameters:  float, xx;
** description: Return log of gamma function of xx for xx > 0.
***********************************************************************/
float Rstream::lngamma(float  xx) const {
  float  x, tmp, ser;
  int j;

  x = xx - 1.0;
  tmp = x + 5.5;
  tmp -= (x + 0.5) * log(tmp);
  ser = 1.0;
  for (j = 0;j < 6;j++) {
         x += 1.0;
         ser += cof[j] / x;
  }
  return (-tmp + log(2.50662827465 * ser));
} // End lngamma


/***********************************************************************
** function:    unit_rate_poisson;
** parameters:  float, mean;
** description: Return a random variable drawn from the poisson
**              distribution with average value mean. The value
**              returned represents the number of unit rate poisson
**              events occurring.
***********************************************************************/
double Rstream::unit_rate_poisson(float mean) {

  static float sq, alxm, g, oldm = -1.0;
  float  em, t, y;

  /*
  ** Can we use direct method
  */
  if (mean < 12.0) {
    if (mean != oldm) {
      oldm = mean;
      g    = exp(-mean);
    }
    em = -1.0;
    t  = 1.0;
    do {
      em += 1.0;
      t *= uniform01();
    } while (t > g);
  }
  /*
  ** If not, use rejection method
  */
  else {
    if (mean != oldm) {
      oldm = mean;
      sq   = sqrt(2.0 * mean);
      alxm = log(mean);
      g    = mean * alxm - lngamma(mean + 1.0);
    }
    do {
      do {
        /*
        ** y is a deviate from a Lorentzian comparison function
        */
        y    = tan (3.141592654 * uniform01());           // pi
        em   = sq * y + mean;
      } while (em < 0.0);
      /*
      ** poisson is integer valued
      */
      em = floor(em);
      /*
      ** t is the ration of the desired distribution to the comparison
      ** function - accept or reject by comparing it to the the uniform
      ** deviate. Factor 0.9 is chosen so that it never exceeds 1.0
      */
      t = 0.9 * (1.0 + y * y) * exp(em * alxm - lngamma(em + 1.0) - g);
    } while (uniform01() > t);
  }
  return em;
} // End poisson


/***********************************************************************
** functions:   initPareto, Pareto
** parameters:  double alpha, double minimum value (b)
** description: Return a random variable drawn from the Pareto
**
**              pdf for Pareto is: f(x)=alpha*b^(alpha)/x^(alpha+1)
**              
**              The Paretian index, alpha, is a measure of dispersion
**              or inequality of distribution.  (For alpha >= 2, we have
**              finite variance, and this variance decreases with
**              aggregation, as governed by the Central Limit Theorem.)
**              
**              The Paretian index is related to the Hurst parameter
**              H by the equivalent equations 
**
**                         H = (3 - alpha)/2,
**              or
**                     alpha = 3 - 2H.
**              
**              Whether the Hurst or alpha parametric form is used,
**              this parameter expresses the degree of LRD of the time
**              series, i.e., the speed of decay of the series'
**              autocorrelation function.
**
**              A Hurst parameter of H = 1/2 corresponds to
**              independence, as does alpha >= 2. Hurst 1/2 < H < 1
**              corresponds to long-term dependence and self-similar
**              character; Paretian index 1 < alpha < 2 indicates
**              long-range dependence. Infinite variance is approached
**              as H --> 1.
**              Typical self similar values are: H=0.9 (alpha=1.2)
**
***********************************************************************/
double Rstream::Pareto(double alpha, double minimum_value)
{
  return minimum_value/pow(uniform01(),1.0/alpha);
} 

/***********************************************************************
** function:    Trimodal_Internet;
** parameters:  none
** description: Generate a discrete random number according to
**    measured packet lengths in real networks (in bytes)
**
** http://www.isoc.org/isoc/conferences/inet/98/proceedings/6g/6g_3.htm
***********************************************************************/
short Rstream::Trimodal_Internet() {
  double r = uniform01();
  if(r<0.6)
    {
      return (short) floor((r/0.6)*5.0+40);
    }
  else if(r<0.75)
    {
      return (short) floor(((r-0.6)/0.15)*509.0+44.0);
    }
  else if(r<0.84)
    {
      return (short) floor(((r-0.75)/0.09)*25.0+552.0);
    }
  else if(r<0.89)
    {
      return (short) floor((r-0.84)*20.0*925.0+576.0);
    }
  else
    {
      return 1500;
    }
} // End uniform_discrete


/***********************************************************************
** function:    double RandomGaussian
** parameters:  (double mean,double stddev)
** description: Generate a random number according to a normal pdf 
** with mean and stdev
**
**    ALGORITHM 712, COLLECTED ALGORITHMS FROM ACM.
**    THIS WORK PUBLISHED IN TRANSACTIONS ON MATHEMATICAL SOFTWARE,
**    VOL. 18, NO. 4, DECEMBER, 1992, PP. 434-435.
**    The function returns a normally distributed pseudo-random number
**    with a given mean and standard devaiation.  Calls are made to a
**    function subprogram which must return independent random
**    numbers uniform in the interval (0,1).
**    The algorithm uses the ratio of uniforms method of A.J. Kinderman
**    and J.F. Monahan augmented with quadratic bounding curves.
***********************************************************************/
double Rstream::RandomGaussian(double mean,double stddev)
{
  double q,u,v,x,y;

  /*
    Generate P = (u,v) uniform in rect. enclosing acceptance region
    Make sure that any random numbers <= 0 are rejected, since
    gaussian() requires uniforms > 0, but RandomUniform() delivers >= 0.*/
  do {
    u = uniform01();
    v = uniform01();

    v = 1.7156 * (v - 0.5);

    /*  Evaluate the quadratic form */
    x = u - 0.449871;
    y = fabs(v) + 0.386595;
    q = x * x + y * (0.19600 * y - 0.25472 * x);

    /* Accept P if inside inner ellipse */
    if (q < 0.27597)
      break;

    /*  Reject P if outside outer ellipse, or outside acceptance region */
  } while ((q > 0.27846) || (v * v > -4.0 * log(u) * u * u));

  /*  Return ratio of P's coordinates as the normal deviate */
  return (mean + stddev * v / u);
}
